import numpy as np

def rgb2luma(arr):
	'''
        convert RGB image to greyscale

        Parameters 
        ----------
        arr : numpy array 
            numpy array containing two spatial dimensions and 3 colour channels

        Returns 
        --------
        arr : numpy array
            2D numpy array containing the luminance conversion of the input 3D 
            colour array
    ''' 
	return np.dot(arr[...,:3], [0.299, 0.587, 0.144])


import os
import nosypy as nosy

def mask(img,t):

    # apply threshold
    highidx         = img>=t
    lowidx          = img<t
    img[highidx]   = 255
    img[lowidx]    = 0

    return img

def maskall(inpath,outpath,t):
    '''
        Generate a mask by applying a threshold
    '''

    if not os.path.exists(outpath): os.mkdir(outpath)
    '''
        Get list of files in directory 
    '''
    filelist = os.listdir(inpath)

    for f in filelist:
        # load image 
        orig        = nosy.image.io.imread(os.path.join(inpath,f))  
        # Apply threshold 
        maskimg     = mask(orig,t)
        # save image
        nosy.image.io.save_image(maskimg,os.path.join(outpath,f))


import requests
import tarfile
import os
import sys
import warnings
from nosypy.image.io import imread
import numpy as np

def downloader(url,localdir):
    r = requests.get(url, allow_redirects=True, stream=True)
    total_length = r.headers.get('content-length')
    output = open(localdir, 'wb')
    if total_length is None: # no content length header
        output.write(r.content)
    else:
        dl = 0
        total_length = int(total_length)
        for data in r.iter_content(total_length/100):
            dl += len(data)
            output.write(data)
            done = int(50 * dl / total_length)
            sys.stdout.write("\r[%s%s]" % ('=' * done, ' ' * (50-done)) )    
            sys.stdout.flush()


def unpack(fname):
    path    = os.path.dirname(os.path.realpath(fname))
    if (fname.endswith("tar.gz")):
        tar = tarfile.open(fname, "r:gz")
        tar.extractall(path=path)
        tar.close()
    elif (fname.endswith("tar")):
        tar = tarfile.open(fname, "r:")
        tar.extractall(path=path)
        tar.close()


def getpaths(path,prefix='',imgrange=None,ext='JPG'):
    '''
        Returns a list of file paths. If the filenames are 
        asigned sequentially, the imgrange=[start,end] list 
        can be used to select the range of images to be 
        added to the output list of paths. Otherwise, all 
        files in path will be added 

        Parameters 
        ---------- 

        Return
        -------
    '''
    if imgrange==None:
        # list of files in directory
        files = os.listdir(path)
        # remove any within an excepted extention 
        files = [x for x in files if any(y in x for y in ['.png','.jpg','.JPG','.tif'])]
        # add path as prefix 
        filepaths = [os.path.join(path,s) for s in files]
    else:
        files = []
        filepaths = []
        for i in range(imgrange[0],imgrange[1]):
            files.append(os.path.join(prefix+str(i)+'.'+ext))
            filepaths.append(os.path.join(path,prefix+str(i)+'.'+ext))
    return files,filepaths

class Set:

    def __init__(self, traindir,validatedir=None):
        '''

        Parameters
        ----------
        traindir : 

        validatedir : 

        Returns
        -------

        '''
        # locate images in path 
        self.train_files,self.train_paths = getpaths(traindir)
        if not validatedir == None: self.val_paths = validatedir

        # determine height and width from first image
        try:self.height, self.width = np.shape(imread(self.train_paths[0]))
        except:self.height, self.width, _ = np.shape(imread(self.train_paths[0]))


class UCSD:

    def __init__(self,localdir='UCDS_dataset'):
        if not os.path.exists(localdir): os.mkdir(localdir)
        if not os.path.exists(os.path.join(localdir,'UCSD_Anomaly_Dataset.v1p2')):
            if not os.path.exists(os.path.join(localdir,'UCSD_Anomaly_Dataset.tar.gz')):
                print('\nDownloading UCDS dataset')
                url = 'http://www.svcl.ucsd.edu/projects/anomaly/UCSD_Anomaly_Dataset.tar.gz'
                try: # add multiple attempts
                    downloader(url,os.path.join(localdir,'UCSD_Anomaly_Dataset.tar.gz'))
                    print ("complete\n")
                except:
                    raise ValueError('Failed to download dataset - the link may be dead :(')
            # MD5 checksum: 5006421b89885f45a6f93b041145f2eb
            print("unpacking dataset...")
            unpack(os.path.join(localdir,'UCSD_Anomaly_Dataset.tar.gz'))
            print("...done")
        
        if not os.path.exists(os.path.join(localdir,'UCSD_Ped1')):
            if not os.path.exists(os.path.join(localdir,'UCSD_Ped1.tar.gz')):
                print('Downloading UCDS groundtruth data')
                gt_url = 'https://hci.iwr.uni-heidelberg.de/sites/default/files/node/files/1109511602/UCSD_Ped1.tar.gz'
                try:
                    downloader(gt_url,os.path.join(localdir,'UCSD_Ped1.tar.gz'))
                    print ("complete\n")
                except:
                    raise ValueError('Failed to download groundtruth data - the link may be dead :(')  
            
            print("unpacking ground truth...")
            unpack(os.path.join(localdir,'UCSD_Ped1.tar.gz'))
            print("...done")

        # set parameters 
        self.height,self.width = 158,238
        # file list 
        ped1dir     = os.path.join(localdir,'UCSD_Anomaly_Dataset.v1p2','UCSDped1','Train')
        # training sets
        self.train001 = getpaths(os.path.join(ped1dir,'Train001'))
        self.train002 = getpaths(os.path.join(ped1dir,'Train002'))
        self.train003 = getpaths(os.path.join(ped1dir,'Train003'))
        self.train004 = getpaths(os.path.join(ped1dir,'Train004'))
        self.train005 = getpaths(os.path.join(ped1dir,'Train005'))
        self.train006 = getpaths(os.path.join(ped1dir,'Train006'))
        self.train007 = getpaths(os.path.join(ped1dir,'Train007'))
        self.train008 = getpaths(os.path.join(ped1dir,'Train008'))
        self.train009 = getpaths(os.path.join(ped1dir,'Train009'))
        self.train009 = getpaths(os.path.join(ped1dir,'Train010'))
        self.train010 = getpaths(os.path.join(ped1dir,'Train011'))
        self.train011 = getpaths(os.path.join(ped1dir,'Train012'))
        self.train012 = getpaths(os.path.join(ped1dir,'Train013'))
        self.train013 = getpaths(os.path.join(ped1dir,'Train014'))
        self.train014 = getpaths(os.path.join(ped1dir,'Train015'))
        self.train016 = getpaths(os.path.join(ped1dir,'Train016'))
        self.train017 = getpaths(os.path.join(ped1dir,'Train017'))
        self.train018 = getpaths(os.path.join(ped1dir,'Train018'))
        self.train019 = getpaths(os.path.join(ped1dir,'Train019'))
        self.train020 = getpaths(os.path.join(ped1dir,'Train020'))
        self.train021 = getpaths(os.path.join(ped1dir,'Train021'))
        self.train022 = getpaths(os.path.join(ped1dir,'Train022'))
        self.train023 = getpaths(os.path.join(ped1dir,'Train023'))
        self.train024 = getpaths(os.path.join(ped1dir,'Train024'))
        self.train025 = getpaths(os.path.join(ped1dir,'Train025'))
        self.train026 = getpaths(os.path.join(ped1dir,'Train026'))
        self.train027 = getpaths(os.path.join(ped1dir,'Train027'))
        self.train028 = getpaths(os.path.join(ped1dir,'Train028'))
        self.train029 = getpaths(os.path.join(ped1dir,'Train029'))
        self.train030 = getpaths(os.path.join(ped1dir,'Train030'))
        self.train031 = getpaths(os.path.join(ped1dir,'Train031'))
        self.train032 = getpaths(os.path.join(ped1dir,'Train032'))
        self.train033 = getpaths(os.path.join(ped1dir,'Train033'))
        self.train034 = getpaths(os.path.join(ped1dir,'Train034'))


# Package modules
from nosypy import autoencoder
from nosypy import data
from nosypy import run 
from nosypy import annotate

# For nested packages 
from nosypy import image 
from nosypy import novelty 
# from nosypy import visualise 

__version__ = '0.0.2'
name = 'nosypy'


import os
import nosypy as nosy
import numpy as np
from skimage import measure

def percentile(inpath,percentile=0.99):
    '''
        Calculate the n percetile
    '''

    
    #Get list of files in directory 
    filelist = os.listdir(inpath)

    # Create array to store frequency of pixel intensities
    hist_arr = np.array(np.zeros(256))
    for f in filelist:
        # load image 
        img        = nosy.image.io.imread(os.path.join(inpath,f))

        img_arr = img.ravel() # is this needed?
        unique, counts  = np.unique(img_arr, return_counts=True)
        for t in range(len(unique)):
            hist_arr[unique[t]] += counts[t]

    total       = float(np.sum(hist_arr))
    cuml        = 0.0
    for i in range(255,0,-1):
        cuml = np.sum(hist_arr[:i])/total
        if cuml <=percentile:
            treshold=i
            break

    return treshold


def meanpercentile(inpath,percentile=0.99):
    '''
        Calculate mean percentile
    '''
    #Get list of files in directory 
    filelist    = os.listdir(inpath)
    nfiles      = len(filelist)
    sumthreshold = 0
    for f in filelist:
        # Create array to store frequency of pixel intensities
        hist_arr = np.array(np.zeros(256))
        # load image 
        img        = nosy.image.io.imread(os.path.join(inpath,f))

        img_arr = img.ravel() # is this needed?
        unique, counts  = np.unique(img_arr, return_counts=True)
        for t in range(len(unique)):
            hist_arr[unique[t]] += counts[t]
        total       = float(np.sum(hist_arr))
        for i in range(255,0,-1):
            cuml = np.sum(hist_arr[:i])/total
            if cuml <=percentile:
                sumthreshold+=i
                break

    return int(np.floor(sumthreshold/float(nfiles)))


def contours(mask):
    '''
        find contours in image mask 
    '''
    return measure.find_contours(mask, 0)

def contour2bbox(contour):
    '''
        Calculate a bounding box from a contour

        Parameters
        ----------
        contour : array_like 
            a single contour

        Returns
        -------
        x and y points of each corner of the bounding box in image
        ymin,ymax,xmin,xmax
    '''
    xmin,xmax       = min(contour[:, 1]), max(contour[:, 1])
    ymin,ymax       = min(contour[:, 0]), max(contour[:, 0])

    return ymin,ymax,xmin,xmax


import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import backend
import os
import numpy as np

class ConvNet:
    '''
        Convolutional Autoencoder Network
    '''
    def __init__(self):

        self.callbacks = []

    def reinitialise(self):
        self.callbacks = []
        backend.clear_session()
        tf.reset_default_graph()

    def tensorboard(self,log='tensorboard/covnet_autoencoder'):
        '''
            Tensorboard
        '''
        self.callbacks.append(keras.callbacks.TensorBoard(log_dir=log,write_graph=True, write_images=True))


    def checkpoints(self,path='training',N=10):
        checkpoint_path = os.path.join(path,"cp-{epoch:04d}.ckpt")

        # Create checkpoint callback
        self.callbacks.append(tf.keras.callbacks.ModelCheckpoint(checkpoint_path,
                                                         save_weights_only=True,
                                                         verbose=1, period=N))

    def model(self,height,width,channels=1):
        '''
            Build the model
            The encoder will consist in a stack of Conv2D and MaxPooling2D layers
            (max pooling being used for spatial down-sampling), while the decoder will
            consist in a stack of Conv2D and UpSampling2D layers.
        '''
        input_img = keras.layers.Input(shape=(height, width, channels))

        x = keras.layers.Conv2D(16, (3, 3), activation='relu', padding='same')(input_img)
        x = keras.layers.MaxPooling2D((2, 2), padding='same')(x)
        x = keras.layers.Conv2D(8, (3, 3), activation='relu', padding='same')(x)
        x = keras.layers.MaxPooling2D((2, 2), padding='same')(x)
        x = keras.layers.Conv2D(8, (3, 3), activation='relu', padding='same')(x)
        encoded = keras.layers.MaxPooling2D((2, 2), padding='same')(x)

        x = keras.layers.Conv2D(8, (3, 3), activation='relu', padding='same')(encoded)
        x = keras.layers.UpSampling2D((2, 2))(x)
        x = keras.layers.Conv2D(8, (3, 3), activation='relu', padding='same')(x)
        x = keras.layers.UpSampling2D((2, 2))(x)
        x = keras.layers.Conv2D(16, (3, 3), activation='relu')(x)
        x = keras.layers.UpSampling2D((2, 2))(x)
        decoded = keras.layers.Conv2D(1, (3, 3), activation='sigmoid', padding='same')(x)

        self.autoencoder = keras.models.Model(input_img, decoded)
        # self.autoencoder.compile(optimizer='adadelta', loss='binary_crossentropy' )
        self.autoencoder.compile(optimizer='Adam', loss='mean_squared_error' )

    def model_summary(self):
        '''
        	Model Summary
        '''
        self.autoencoder.summary()

    def train(self,X,epoch=100,batch_size=256):
        '''
            Train the model
            X samples should be in the range [0,1]
        '''

        self.autoencoder.fit(X, X,
                        epochs=epoch,
                        batch_size=batch_size,
                        shuffle=True,
                        # validation_data=(test_images, test_images),
                        callbacks=self.callbacks)


    def save_model(self, path='model.h5'):
        self.autoencoder.save_weights(path)

    def load_model(self,path='model.h5'):
        self.autoencoder.load_weights(path)


    def maxsse(self, height, width):
        '''
            calculate the maximum sse posible given the width and height of samples
        '''
        x = np.array(np.zeros((height,width)))
        y = np.array(np.ones((height,width)))
        return np.sum(np.power(np.subtract(x,y),2))

    def sse(self, x):
        '''
            return sum of squared error between input x and output x'
        '''
        decoded = self.autoencoder.predict(x[np.newaxis,:,:,:])
        return np.sum(np.power(np.subtract(x,decoded),2))

    def decode(self, X):
        '''
            return sum of squared error between input x and output x'
            X [samples, height,width, 1]
        '''
        return self.autoencoder.predict(X)


'''
    Run NATI on the UCSD Anomaly Detection Dataset
'''
try:
    import nosypy as nosy
except:
    from context import nosypy as nosy

import os
'''
    Create directories for data
'''
data    = 'ucsd_data'
output  = 'nosypy_output' 
if not os.path.exists(data):os.mkdir(data)
if not os.path.exists(output):os.mkdir(output)
    
'''
    Download dataset
'''
ds      = nosy.data.UCSD(localdir=data) 

'''
    The UCSD dataset has 34 Training sets in ped1 and the file paths are given in 
    UCDS.train001 to UCDS.train034  

    Train NATI
'''
pheight, pwidth = 28,28
jtiles, itiles  = 5,7

nosy.run.train_models(ds.train001,output,ds.height,ds.width,pheight, pwidth, jtiles, itiles,epoch=1)




def contours(img,c):
    '''
        Draw contours on image

    '''
    raise NotImplementedError


def rect(img,corners, color='r',linewidth=3):
    '''
        Draw rectangle on image 

        Add limits according to image dimensions
        Parameters 
        ----------
        corners : list
            ymin,ymax,xmin,xmax
        linewidth : int 
            number of pixels the lines od the rectangle cover. Should 
            be odd.
        Returns 
        -------
    '''
    #get colour in rgb 
    colrgb = [255,0,0] # add a colour function selector here
    # line overlap
    olap        = (linewidth-1)/2
    # Draw top horizontal line 
    img[corners[1]-olap:corners[1]+olap+1,corners[2]:corners[3]] = colrgb
    # Draw bottom horizontal line
    img[corners[0]-olap:corners[0]+olap+1,corners[2]:corners[3]] = colrgb
    # Draw  left vertical line 
    img[corners[0]:corners[1],corners[2]-olap:corners[2]+olap+1] = colrgb
    # Draw  left vertical line 
    img[corners[0]:corners[1],corners[3]-olap:corners[3]+olap+1] = colrgb

    return img

def colourstr2rgb(colorstr):
    '''


    '''
    


import numpy as np
import os 
import nosypy as nosy
from scipy import ndimage

def lincoeff(filterwidth=31):
    '''
        Generate a 1D array of linear filter coefficients
        filterwidth should be odd, otherwise filterwidth-1 will be taken

        Parameters
        ----------

    '''
    olap = int((filterwidth+1)/2)
    return np.lib.pad(np.linspace(1,3,olap), (0,olap-1),'reflect')

def linear2D(img,filterwidth=31):
    '''
        Apply a 2D linear filter to image 

        Parameters
        ----------
        img : array_like
    '''
    kernel = kernel2D(lincoeff(filterwidth=filterwidth))
    return ndimage.convolve(img,kernel)


def linear2Dall(inpath,outpath,filterwidth=31):
    '''
        Use a 2D linear filter on each image in inpath
    '''
    kernel = kernel2D(lincoeff(filterwidth=filterwidth))
    if not os.path.exists(outpath): os.mkdir(outpath)
    
    '''
        Get list of files in directory 
    '''
    filelist = os.listdir(inpath)

    for f in filelist:
        # load image 
        orig        = nosy.image.io.imread(os.path.join(inpath,f))
        # resize image 
        filtered    = ndimage.convolve(orig,kernel)
        # save image 
        nosy.image.io.save_image(filtered,os.path.join(outpath,f))


def kernel2D(coeffs):
    '''
        Create a 2D normalised kernel from a 1D array

        Parameters
        ----------
        coeffs : numpy array
            1D numpy array of filter coefficients 

    '''
    size = len(coeffs)
    Unit = np.array(np.ones((size,size)),dtype=float)
    filtx = [coeffs]
    Kx = np.multiply(filtx,Unit)
    filty = [coeffs]
    Ky = np.multiply(Unit,np.transpose(filty))
    Kernel = np.multiply(Kx,Ky)
    Normalise = np.sum(Kernel)
    return np.divide(Kernel,Normalise)


import nosypy as nosy
import numpy as np
import time
import os


def detect(filepaths,outpath,height,width,pheight, pwidth, jtiles, itiles):
    '''


    '''
    
    '''
        Define tiles
    '''
    tiles = nosy.image.tile.create(height,width,jtiles,itiles, overlaps=False)
    ntiles  = len(tiles)
    print(ntiles, " tiles")


    '''
        Build model
    '''
    aen     = nosy.autoencoder.ConvNet()
    aen.model(pheight,pwidth)
    aen.model_summary()

    '''
        Output directories
    '''
    noveltymaps             = os.path.join(outpath, 'novelty')
    novmap_npy_path         = os.path.join(outpath, 'novelty', 'numpy_error')
    novmap_setscale_path    = os.path.join(outpath, 'novelty', 'set_norm')
    if not os.path.exists(noveltymaps): os.mkdir(noveltymaps)
    if not os.path.exists(novmap_npy_path): os.mkdir(novmap_npy_path)
    if not os.path.exists(novmap_setscale_path): os.mkdir(novmap_setscale_path)

    setmax = 0.0    # maximum sse of set
    for imgidx in range(len(filepaths)):
        '''
            Extract filename from path 
        '''
        filename = os.path.basename(filepaths[imgidx])
        print ("Image: ", filename)
        # load image
        img = np.divide(nosy.image.io.imread(filepaths[imgidx]).astype('float'),255.0)

        '''
        # Initialise Novety map for image
        '''
        Nmap = np.array(np.zeros((height,width)))
        Nweights = np.array(np.zeros((height,width)))

        for tile_id in range(ntiles):
            tile            = tiles[tile_id]

            '''
                X patch array
                need to create in loop because tile dimensions can change near borders
            '''
            nframepatches   = int((tile[1]-tile[0]+1) * (tile[3]-tile[2]+1))
            # print (nframepatches, ' patches in tile ', str(tile_id))

            # Load Tile Model
            aen.load_model(path=os.path.join(outpath,'models','aen_tile' + str(tile_id) + '_model.h5'))

            # Calculate difference between X and X' and create novety map
            Nmap,Nweights = nosy.novelty.patch.error(img,pheight,pwidth,tile,aen,nframepatches,Nmap=Nmap, Nweights=Nweights)


        # Normalise Novelty map and save
        Nmap        = np.divide(Nmap,Nweights)
        imgmax = np.max(Nmap) # for tracking maximum value in set
        if imgmax > setmax:
            setmax = imgmax
        # andect.ip.save_image(Nmap,os.path.join(novmap_path, fileprefix+str(img_range[0] + img_id)+'.png'))
        np.save(os.path.join(novmap_npy_path, filename +'.npy'),Nmap)

    '''
        Normalise/scale images according to max sse of entire set
    '''
    for imgidx in range(len(filepaths)):
        '''
            Extract filename from path 
        '''
        filename = os.path.basename(filepaths[imgidx])

        img = np.load(os.path.join(novmap_npy_path, filename +'.npy'))
        nosy.image.io.save_image(np.multiply(img,255.0/setmax),os.path.join(novmap_setscale_path, filename +'.png'))

    '''
        Remove temporary numpy files 
    '''
    

def train_models(filepaths,outpath,height,width,pheight, pwidth, jtiles, itiles,epoch=20):
    '''

    Parameters
    ----------
    pheight :

    pwidth : 

    jtiles :

    itiles :


    Returns
    -------

    '''
    def train(aen,tiles,tilelist,model_dir,filepaths,pheight,pwidth,epoch=20):

        for tile_id in tilelist:
            start_time = time.time()
            outputmodel = os.path.join(model_dir,'aen_tile' + str(tile_id) + '_model.h5')
            if os.path.exists(outputmodel): continue

            '''
                Callbacks
            '''
            checkpoint_path = os.path.join(model_dir,'tile' + str(tile_id) + '_training')
            if not os.path.exists(checkpoint_path): os.mkdir(checkpoint_path)
            tensorboard_path = os.path.join(model_dir,'tile' + str(tile_id) + '_tensorboard/covnet_autoencoder')
            aen.reinitialise()
            aen.tensorboard(log=tensorboard_path)
            aen.checkpoints(path=checkpoint_path,N=10)

            aen.model(pheight,pwidth)
            aen.model_summary()

            '''
                Extract patches in tile and add to list
            '''
            tile            = tiles[tile_id]
            print("tile: ", tile_id)
            for i in range(len(filepaths)):
                # load image and convert to range [0,1]
                img = np.divide(nosy.image.io.imread(filepaths[i]).astype('float'),255.0)
                X[i*nframepatches:(i+1)*nframepatches,...] = nosy.image.tile.extract_patches(img,pheight,pwidth,tile)


            aen.train(X,epoch=epoch)
            aen.save_model(path=outputmodel)
            elapsed_time = time.time() - start_time
            print ("Time elapsed: ", elapsed_time, " seconds")
            print ("Time elapsed: ", elapsed_time/60.0, " minutes")


    # create working directory
    if not os.path.exists(outpath): os.mkdir(outpath)

    model_dir               = os.path.join(outpath,'models')
    if not os.path.exists(model_dir): os.mkdir(model_dir)

    # define tiles
    tiles = nosy.image.tile.create(height,width,jtiles,itiles, overlaps=False)
    ntiles  = len(tiles)
    print(ntiles, " tiles")
    # Create Model Class
    aen     = nosy.autoencoder.ConvNet()
    nimages         = len(filepaths)
    
    #### Train tiles not on the right or bottom border ###
    nframepatches   = int((tiles[0][1]-tiles[0][0]+1) * (tiles[0][3]-tiles[0][2]+1))
    npatches        = nframepatches*nimages
    print (npatches, ' patches')

    # array for storiing patches
    X = np.array(np.zeros((npatches, pheight, pwidth, 1)))

    tilelist = []
    for i in range(jtiles-1):
        tilelist.extend(np.arange((i*itiles),(((i+1)*itiles)-1)))
    train(aen,tiles,tilelist,model_dir,filepaths,pheight,pwidth,epoch=epoch)

    #### Train tiles on the right border ###
    nframepatches   = int((tiles[itiles][1]-tiles[itiles][0]+1) * (tiles[itiles][3]-tiles[itiles][2]+1))
    npatches        = nframepatches*nimages

    # array for storiing patches
    X = np.array(np.zeros((npatches, pheight, pwidth, 1)))

    tilelist = []
    for i in range(jtiles-1):
        tilelist.append(i*itiles)
    # train(aen,tiles,tilelist,model_dir,filepaths,pheight,pwidth,epoch=epoch)

    #### Train tiles on the bottom right tile ###
    nframepatches   = int((tiles[(itiles*jtiles)-1][1]-tiles[(itiles*jtiles)-1][0]+1) * (tiles[(itiles*jtiles)-1][3]-tiles[(itiles*jtiles)-1][2]+1))
    npatches        = nframepatches*nimages

    # array for storiing patches
    X = np.array(np.zeros((npatches, pheight, pwidth, 1)))

    tilelist = []
    for i in range(jtiles-1):
        tilelist.append(i*itiles)
    # train(aen,tiles,tilelist,model_dir,filepaths,pheight,pwidth,epoch=epoch)




# if __name__=='__main__':
#     detection(filepaths,outpath,height,width,pheight, pwidth, jtiles, itiles)


import numpy as np

def error(img,pheight,pwidth,tile,model,nframepatches,Nmap=None, Nweights=None, MODE=1):
    '''
        Adds to a novety map and normalisation map (Nweights).

        Compare x with x' and
        MODE 1:
            add difference to the centre 2x2 block of the patch.
            Each addition in Nmap accumlates in the Nweights array, so that values
            can be normalised after.

        MODE 2:
            pixel difference

        The centre of even patches is considered the bottom right pixel of the centre
        2x2 block. Overlap for even patches is the largest overlap from this pixel
        (to the left and to the top).

    '''
    try:height,width    = np.shape(img)
    except:height,width,_    = np.shape(img)


    if pheight%2==0:# even
        jolap=int(pheight/2)
        jrange=[int(jolap+tile[0]),int(tile[1]+jolap+1)] # of padded image
    else: # odd
        # jolap=int((pheight-1)/2)
        # jrange = [jolap,height-jolap]
        raise NotImplementedError

    if pwidth%2==0:# even
        iolap = int(pwidth/2)
        irange = [int(iolap+tile[2]),int(tile[3]+iolap+1)]
    else:
        # iolap=int((pwidth-1)/2)
        # irange = [iolap,width-iolap]
        raise NotImplementedError

    # pad image to account for overlapping patches
    padded = np.pad(img, ((jolap,iolap),(jolap,iolap)), 'reflect')[:,:,np.newaxis] # (up,left),(down,right)

    X = np.array(np.zeros((nframepatches,pheight,pwidth,1)))
    pidx = 0
    for j in range(jrange[0],jrange[1]):
        for i in range(irange[0],irange[1]):
            X[pidx,...] = padded[j-jolap:j+jolap,i-iolap:i+iolap] # for even pheight and pwidth
            pidx+=1

    # Run on all patches
    decoded = model.decode(X) # performing this on all samples is faster
    idx=0

    if MODE == 1: # Store error for entire patch in centre 2x2 block
        for j in range(jrange[0],jrange[1]):
            for i in range(irange[0],irange[1]):
                error = np.sum(np.power(np.subtract(X[idx,...],decoded[idx,...]),2))
                # add to centre 2x2
                Nmap[           max(j-jolap-1,0):min(j-jolap+1,height),
                                max(i-iolap-1,0):min(i-iolap+1,width)] =  np.add(Nmap[max(j-jolap-1,0):min(j-jolap+1,height),
                                                                                max(i-iolap-1,0):min(i-iolap+1,width)],error) # not padded

                Nweights[       max(j-jolap-1,0):min(j-jolap+1,height),
                                max(i-iolap-1,0):min(i-iolap+1,width)] =  np.add(Nweights[max(j-jolap-1,0):min(j-jolap+1,height),
                                                                                max(i-iolap-1,0):min(i-iolap+1,width)],1.0) # not padded
                idx+=1

    elif MODE == 2: # store pixel-wise error in corresponding pixel
        for j in range(jrange[0],jrange[1]):
            for i in range(irange[0],irange[1]):
                # print(j,i)
                error = np.power(np.subtract(   X[idx,  max(0,pheight-j-1):   min(pheight,height-j+pheight-1),
                                                        max(0,pwidth-i-1):    min(pwidth,width-i+pwidth-1)],
                                                decoded[idx,max(0,pheight-j-1):   min(pheight,height-j+pheight-1),
                                                            max(0,pwidth-i-1):    min(pwidth,width-i+pwidth-1)]),2)[:,:,0]
                # print(np.shape(error))
                # print(np.shape(Nmap[  max(j-(2*jolap)+1,0):min(j+1,height),
                                                                    # max(i-(2*iolap)+1,0):min(i+1,width)]),i+1,width)
                Nmap[           max(j-(2*jolap)+1,0):min(j+1,height),
                                max(i-(2*iolap)+1,0):min(i+1,width)] =  np.add(Nmap[    max(j-(2*jolap)+1,0):min(j+1,height),
                                                                                        max(i-(2*iolap)+1,0):min(i+1,width)],error) # not padded

                Nweights[       max(j-(2*jolap)+1,0):min(j+1,height),
                                max(i-(2*iolap)+1,0):min(i+1,width)] =  np.add(Nweights[  max(j-(2*jolap)+1,0):min(j+1,height),
                                                                                            max(i-(2*iolap)+1,0):min(i+1,width)],1.0) # not padded
                idx+=1
    else:
        raise ValueError("Invalid mode in data.novelty_map()")

    return Nmap,Nweights


import numpy as np
import imageio

def imread(path):
    '''
        Use imageio library to load an image as a numpy array

        Parameters
        ----------
        path : string 
            path to the image file

        Returns
        ------- 
        image : array_like 
    '''
    return imageio.imread(path)


def save_image(arr,output):
    '''
        Use imageio library to save an image after clipping 
        the input between the range [0,255] and converting
        to unsigned 8-bit int

        Parameters
        ----------
        arr : numpy array 
            image array with either 1 (greyscale) or 3 (RGB)
            channels
        output : string 
            path to output

        Returns
        -------
        None
    '''
    imageio.imwrite(output,np.clip(arr,0,255).astype('uint8'))


def save_npy(arr,output):
    raise NotImplementedError


#!/usr/bin/env python
""" NosyPy is a self supervised novelty detection package built for time lapse imagery using Python.

    - It is built ontop of Tensorflow 
    - It uses a self supervised learning approach, removing the need for manual annotation
    - It identifies areas in each image which likely contain an object, or objects, of interest   
"""

from distutils.core import setup
from setuptools import find_packages

DOCLINES = (__doc__ or '').split("\n")

setup(name='nosypy',
      version='0.0.2',
      description=DOCLINES[0],
      long_description="\n".join(DOCLINES[0:]),
      url='http://github.com/brett-hosking/nosypy',
      license='MIT',
      author='brett hosking',
      author_email='wilski@noc.ac.uk',
      install_requires=[
                "tensorflow>=1.12.0",
                "requests>=2.21.0",
                "imageio>=2.5.0",
                "scipy>=1.2.0",
                "scikit-image>=0.14.2"
                ],
      packages=find_packages()
      )



from scipy import ndimage
import nosypy as nosy 
import os 

def bicubic(arr,scale):
    '''
        User scipy library to resize image array 
        using bicubic interpolation

        Parameters 
        ----------
        arr : numpy array 
            2D array or 3D (RGB) image array
        scale : int or float 
            scale factor. <1 decreases spatial resolution, >1
            increases spatial resolution

        Returns 
        -------
        resized array
    '''
    return ndimage.zoom(arr.astype(float), scale, order=3)


def allbicubic(inpath,outpath,scale):
    '''
        Scale all images in path 
        Move this to another module
    '''
    if not os.path.exists(outpath): os.mkdir(outpath)
    
    '''
        Get list of files in directory 
    '''
    filelist = os.listdir(inpath)

    for f in filelist:
        # load image 
        orig        = nosy.image.io.imread(os.path.join(inpath,f))
        # resize image 
        scaled      = nosy.image.resize.bicubic(orig, scale)
        # save image 
        nosy.image.io.save_image(scaled,os.path.join(outpath,f))


import numpy as np 


def regions(img,tiles,colour=[255,0,0],linewidth=3):
    '''
        Visualise the tile regions as produced by data.split_frame()


        Parameters
        ----------
        img : numpy array
            image

        tiles : array
            numpy array of size (tiles,4), [[tile index],jstart,jend,istart,iend]]

        Returns
        -------
    '''

    try:
        height,width,_ = np.shape(img)
    except:
        height,width = np.shape(img)
        print ("viz: ", height,width)
        img2 = np.array(np.zeros((height,width,3)))
        img2[:,:,0] = img
        img2[:,:,1] = img
        img2[:,:,2] = img
        img         = img2
    if np.shape(colour) == (len(tiles),3):
        for t in range(np.shape(tiles)[0]):
            # Draw tile t
            # vertical left line
            img[int(tiles[t][0]):int(tiles[t][1]),int(max(tiles[t][2]-(linewidth/2),0)):int(min(tiles[t][2]+(linewidth/2),width))]           = colour[t]
            # vertical right line
            img[int(tiles[t][0]):int(tiles[t][1]),int(max(tiles[t][3]-(linewidth/2),0)):int(min(tiles[t][3]+(linewidth/2),width))]           = colour[t]
            # horizontal top line
            img[int(max(tiles[t][0]-(linewidth/2),0)):int(min(tiles[t][0]+(linewidth/2),height)),int(tiles[t][2]):int(tiles[t][3])]          = colour[t]
            # horizontal bottom line
            img[int(max(tiles[t][1]-(linewidth/2),0)):int(min(tiles[t][1]+(linewidth/2),height)),int(tiles[t][2]):int(tiles[t][3])]          = colour[t]

    else:
        colour=[255,0,0]
        for t in range(np.shape(tiles)[0]):
            # Draw tile t
            # vertical left line
            img[int(tiles[t][0]):int(tiles[t][1]),int(max(tiles[t][2]-(linewidth/2),0)):int(min(tiles[t][2]+(linewidth/2),width))]           = colour
            # vertical right line
            img[int(tiles[t][0]):int(tiles[t][1]),int(max(tiles[t][3]-(linewidth/2),0)):int(min(tiles[t][3]+(linewidth/2),width))]           = colour
            # horizontal top line
            img[int(max(tiles[t][0]-(linewidth/2),0)):int(min(tiles[t][0]+(linewidth/2),height)),int(tiles[t][2]):int(tiles[t][3])]          = colour
            # horizontal bottom line
            img[int(max(tiles[t][1]-(linewidth/2),0)):int(min(tiles[t][1]+(linewidth/2),height)),int(tiles[t][2]):int(tiles[t][3])]          = colour
    return img



import numpy as np

def create(height,width,jtiles,itiles, overlaps=False):
    '''
        Split frame into multiple tiles and return the tile coordinates

        Parameters
        ----------
        height : int
            height of frame
        width : int
            width of frame
        jtiles : int
            the number of tiles to split the frame into in the vertical dimension
        itiles : int
            the number of tiles to split the frame into in the horizontal dimension
        overlaps : boolean
            if True: create addition tiles of equal size that are centred at the
            original tile cross boundaries

        Returns
        -------
        array of size (tiles,4) which includes the tiles' starting j index, ending
        j index, i starting index, and i ending index, respectively.
    '''
    tileheight, tilewidth = np.ceil(height/jtiles),np.ceil(width/itiles)
    ntiles          = jtiles*itiles
    if overlaps:
        # number of cross boundaries, number of overlap tiles
        ontiles     = (jtiles-1)*(itiles-1)
        ntiles+=ontiles
    tiles           = np.array(np.zeros((ntiles,4)))
    tile_id         = 0
    for j in range(jtiles): # if last tile expand to edge
        for i in range(itiles):

            tiles[tile_id] = [  int(j*tileheight),
                                int(min((j*tileheight)+tileheight,height)),
                                int(i*tilewidth),
                                int(min((i*tilewidth)+tilewidth,width))] # jstart,jend,istart,iend

            if i==itiles-1: # if last tile expand to edge
                tiles[tile_id][3] = int(width)
            if j==jtiles-1: # if last tile expand to edge
                tiles[tile_id][1] = int(height)


            tile_id+=1

    if overlaps:
        # add overlapping tiles
        for j in range(jtiles-1):
            for i in range(itiles-1):
                jcentre,icentre     =  int(min((j*tileheight)+tileheight,height)),int(min((i*tilewidth)+tilewidth,width))
                tiles[tile_id]      =  [    int(jcentre-(tileheight/2)),
                                            int(jcentre+(tileheight/2)),
                                            int(icentre-(tilewidth/2)),
                                            int(icentre+(tilewidth/2))] # jstart,jend,istart,iend
                tile_id+=1

    return tiles


def extract_patches(img,pheight,pwidth, tile):
    '''
        The centre of even patches is considered the bottom right pixel of the centre
        2x2 block. Overlap for even patches is the largest overlap from this pixel
        (to the left and to the top).

        Parameters
        ----------

        tile : list of size (4)
            contains [starting j index, ending j index, i starting index, and i ending index]

    '''
    try:height,width    = np.shape(img)
    except:height,width,_    = np.shape(img)
    if pheight%2==0:# even
        jolap=int(pheight/2)
        jrange=[int(jolap+tile[0]),int(tile[1]+jolap+1)] # of padded image
    else: # odd
        # jolap=int((pheight-1)/2)
        # jrange = [jolap,height-jolap]
        raise NotImplementedError

    if pwidth%2==0:# even
        iolap = int(pwidth/2)
        irange = [int(iolap+tile[2]),int(tile[3]+iolap+1)]
    else:
        # iolap=int((pwidth-1)/2)
        # irange = [iolap,width-iolap]
        raise NotImplementedError

    # pad image to account for overlapping patches
    padded = np.pad(img, ((jolap,iolap),(jolap,iolap)), 'reflect')[:,:,np.newaxis] # (up,left),(down,right)
    patches = []
    for j in range(jrange[0],jrange[1]):
        for i in range(irange[0],irange[1]):
            patches.append(padded[j-jolap:j+jolap,i-iolap:i+iolap]) # for even pheight and pwidth

    return patches
